# Copyright 2011 OpenStack Foundation
# Copyright 2013 Mirantis, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from troveclient import base


class Datastore(base.Resource):

    def __repr__(self):
        return "<Datastore: %s>" % self.name


class DatastoreVersion(base.Resource):

    def __repr__(self):
        return "<DatastoreVersion: %s>" % self.name

    def update(self, visibility=None):
        """Change something in a datastore version."""
        self.manager.update(self.datastore, self.id, visibility)


class DatastoreVersionMember(base.Resource):

    def __repr__(self):
        return "<DatastoreVersionMember: %s>" % self.id


class Datastores(base.ManagerWithFind):
    """Manage :class:`Datastore` resources."""
    resource_class = Datastore

    def __repr__(self):
        return "<Datastore Manager at %s>" % id(self)

    def list(self, limit=None, marker=None):
        """Get a list of all datastores.

        :rtype: list of :class:`Datastore`.
        """
        return self._paginated("/datastores", "datastores", limit, marker)

    def get(self, datastore):
        """Get a specific datastore.

        :rtype: :class:`Datastore`
        """
        return self._get("/datastores/%s" % base.getid(datastore),
                         "datastore")

    def delete(self, datastore):
        """Delete a specific datastore."""
        return self._delete("/datastores/%s" % base.getid(datastore))


class DatastoreVersions(base.ManagerWithFind):
    """Manage :class:`DatastoreVersion` resources."""
    resource_class = DatastoreVersion

    def __repr__(self):
        return "<DatastoreVersions Manager at %s>" % id(self)

    def list(self, datastore, limit=None, marker=None):
        """Get a list of all datastore versions.

        :rtype: list of :class:`DatastoreVersion`.
        """
        return self._paginated("/datastores/%s/versions" % datastore,
                               "versions", limit, marker)

    def get(self, datastore, datastore_version):
        """Get a specific datastore version.

        :rtype: :class:`DatastoreVersion`
        """
        return self._get("/datastores/%s/versions/%s" %
                         (datastore, base.getid(datastore_version)),
                         "version")

    def get_by_uuid(self, datastore_version):
        """Get a specific datastore version.

        :rtype: :class:`DatastoreVersion`
        """
        return self._get("/datastores/versions/%s" %
                         base.getid(datastore_version),
                         "version")

    def update(self, datastore, datastore_version, visibility):
        """Update a specific datastore version."""
        body = {
            "datastore_version": {
            }
        }
        if visibility is not None:
            body["datastore_version"]["visibility"] = visibility

        url = ("/mgmt/datastores/%s/versions/%s" %
               (datastore, datastore_version))
        return self._update(url, body=body)


class DatastoreVersionMembers(base.ManagerWithFind):
    """Manage :class:`DatastoreVersionMember` resources."""
    resource_class = DatastoreVersionMember

    def __repr__(self):
        return "<DatastoreVersionMembers Manager at %s>" % id(self)

    def add(self, datastore, datastore_version, tenant):
        """Add a member to a datastore version."""
        body = {"member": tenant}
        return self._create("/mgmt/datastores/%s/versions/%s/members" %
                            (datastore, datastore_version),
                            body, "datastore_version_member")

    def delete(self, datastore, datastore_version, member_id):
        """Delete a member from a datastore version."""
        return self._delete("/mgmt/datastores/%s/versions/%s/members/%s" %
                            (datastore, datastore_version, member_id))

    def list(self, datastore, datastore_version, limit=None, marker=None):
        """List members of datastore version."""
        return self._list("/mgmt/datastores/%s/versions/%s/members" %
                          (datastore, datastore_version),
                          "datastore_version_members", limit, marker)

    def get(self, datastore, datastore_version, member_id):
        """Get a datastore version member."""
        return self._get("/mgmt/datastores/%s/versions/%s/members/%s" %
                         (datastore, datastore_version, member_id),
                         "datastore_version_member")

    def get_by_tenant(self, datastore, tenant, limit=None, marker=None):
        """List members by tenant id."""
        return self._list("/mgmt/datastores/%s/versions/members/%s" %
                          (datastore, tenant), "datastore_version_members",
                          limit, marker)



# Copyright 2011 OpenStack Foundation
# Copyright 2013 Rackspace Hosting
# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from troveclient import base
from troveclient import common
from troveclient import exceptions
from troveclient import utils
from troveclient.v1 import modules as core_modules

from swiftclient import client as swift_client

REBOOT_SOFT = 'SOFT'
REBOOT_HARD = 'HARD'


class Instance(base.Resource):
    """An Instance is an opaque instance used to store Database instances."""

    def list_databases(self):
        return self.manager.databases.list(self)

    def delete(self):
        """Delete the instance."""
        self.manager.delete(self)

    def force_delete(self):
        """Force delete the instance"""
        self.manager.reset_status(self)
        self.manager.delete(self)

    def restart(self):
        """Restart the database instance."""
        self.manager.restart(self.id)

    def detach_replica(self):
        """Stops the replica database from being replicated to."""
        self.manager.edit(self.id, detach_replica_source=True)


class DatastoreLog(base.Resource):
    """A DatastoreLog is a log on the database guest instance."""

    def __repr__(self):
        return "<DatastoreLog: %s>" % self.name


class Instances(base.ManagerWithFind):
    """Manage :class:`Instance` resources."""
    resource_class = Instance

    def _get_swift_client(self):
        if hasattr(self.api.client, 'auth'):
            auth_url = self.api.client.auth.auth_url
            user = self.api.client.auth._username
            key = self.api.client.auth._password
            tenant_name = self.api.client.auth._project_name
            auth_version = "3.0"
        else:
            auth_url = self.api.client.auth_url
            user = self.api.client.username
            key = self.api.client.password
            tenant_name = self.api.client.tenant
            auth_version = "2.0"

        # remove '/tokens' from the end of auth_url so it works for swift
        token_str = "/tokens"
        if auth_url.endswith(token_str):
            auth_url = auth_url[:-len(token_str)]
        region_name = self.api.client.region_name
        os_options = {'tenant_name': tenant_name, 'region_name': region_name}

        return swift_client.Connection(
            auth_url, user, key, auth_version=auth_version,
            os_options=os_options)

    def create(self, name, flavor_id, volume=None, databases=None, users=None,
               restorePoint=None, availability_zone=None, datastore=None,
               datastore_version=None, nics=None, configuration=None,
               replica_of=None, replica_count=None, modules=None,
               locality=None, region_name=None, access=None, **kwargs):
        """Create (boot) a new instance."""

        body = {"instance": {
            "name": name,
            "flavorRef": flavor_id
        }}
        datastore_obj = {}
        if volume:
            body["instance"]["volume"] = volume
        if databases:
            body["instance"]["databases"] = databases
        if users:
            body["instance"]["users"] = users
        if restorePoint:
            body["instance"]["restorePoint"] = restorePoint
        if availability_zone:
            body["instance"]["availability_zone"] = availability_zone
        if datastore:
            datastore_obj["type"] = datastore
        if datastore_version:
            datastore_obj["version"] = datastore_version
        if datastore_obj:
            body["instance"]["datastore"] = datastore_obj
        if nics:
            body["instance"]["nics"] = nics
        if configuration:
            body["instance"]["configuration"] = base.getid(configuration)
        if replica_of:
            body["instance"]["replica_of"] = base.getid(replica_of)
        if replica_count:
            body["instance"]["replica_count"] = replica_count
        if modules:
            body["instance"]["modules"] = self._get_module_list(modules)
        if locality:
            body["instance"]["locality"] = locality
        if region_name:
            body["instance"]["region_name"] = region_name
        if access:
            body["instance"]["access"] = access

        return self._create("/instances", body, "instance")

    def modify(self, instance, configuration=None):
        body = {
            "instance": {
            }
        }
        if configuration is not None:
            body["instance"]["configuration"] = base.getid(configuration)
        url = "/instances/%s" % base.getid(instance)
        resp, body = self.api.client.put(url, body=body)
        common.check_for_exceptions(resp, body, url)

    def edit(self, instance, configuration=None, name=None,
             detach_replica_source=False, remove_configuration=False):
        body = {
            "instance": {
            }
        }
        if configuration and remove_configuration:
            raise Exception("Cannot attach and detach configuration "
                            "simultaneously.")
        if remove_configuration:
            body["instance"]["configuration"] = None
        if configuration is not None:
            body["instance"]["configuration"] = base.getid(configuration)
        if name is not None:
            body["instance"]["name"] = name
        if detach_replica_source:
            body["instance"]["replica_of"] = None

        url = "/instances/%s" % base.getid(instance)
        resp, body = self.api.client.patch(url, body=body)
        common.check_for_exceptions(resp, body, url)

    def upgrade(self, instance, datastore_version):
        """Upgrades an instance with a new datastore version."""
        body = {
            "instance": {
                "datastore_version": datastore_version
            }
        }

        url = "/instances/%s" % base.getid(instance)
        resp, body = self.api.client.patch(url, body=body)
        common.check_for_exceptions(resp, body, url)

    def list(self, limit=None, marker=None, include_clustered=False,
             detailed=False):
        """Get a list of all instances.

        :rtype: list of :class:`Instance`.
        """
        detail = "/detail" if detailed else ""
        url = "/instances%s" % detail
        return self._paginated(url, "instances", limit, marker,
                               {"include_clustered": include_clustered})

    def get(self, instance):
        """Get a specific instances.

        :rtype: :class:`Instance`
        """
        return self._get("/instances/%s" % base.getid(instance),
                         "instance")

    def backups(self, instance, limit=None, marker=None):
        """Get the list of backups for a specific instance.

        :param instance: instance for which to list backups
        :param limit: max items to return
        :param marker: marker start point
        :rtype: list of :class:`Backups`.
        """
        url = "/instances/%s/backups" % base.getid(instance)
        return self._paginated(url, "backups", limit, marker)

    def delete(self, instance):
        """Delete the specified instance.

        :param instance: A reference to the instance to delete
        """
        url = "/instances/%s" % base.getid(instance)
        resp, body = self.api.client.delete(url)
        common.check_for_exceptions(resp, body, url)

    def reset_status(self, instance):
        """Reset the status of an instance.

        :param instance: A reference to the instance
        """
        body = {'reset_status': {}}
        self._action(instance, body)

    def force_delete(self, instance):
        """Force delete the specified instance.

        :param instance: A reference to the instance to force delete
        """
        self.reset_status(instance)
        self.delete(instance)

    def _action(self, instance, body):
        """Perform a server "action" -- reboot/rebuild/resize/etc."""
        url = "/instances/%s/action" % base.getid(instance)
        resp, body = self.api.client.post(url, body=body)
        common.check_for_exceptions(resp, body, url)
        if body:
            return self.resource_class(self, body, loaded=True)
        return body

    def resize_volume(self, instance, volume_size):
        """Resize the volume on an existing instances."""
        body = {"resize": {"volume": {"size": volume_size}}}
        self._action(instance, body)

    def resize_instance(self, instance, flavor_id):
        """Resizes an instance with a new flavor."""
        body = {"resize": {"flavorRef": flavor_id}}
        self._action(instance, body)

    def restart(self, instance):
        """Restart the database instance.

        :param instance: The :class:`Instance` (or its ID) of the database
                         instance to restart.
        """
        body = {'restart': {}}
        self._action(instance, body)

    def configuration(self, instance):
        """Get a configuration on instances.

        :rtype: :class:`Instance`
        """
        return self._get("/instances/%s/configuration" % base.getid(instance),
                         "instance")

    def promote_to_replica_source(self, instance):
        """Promote a replica to be the new replica_source of its set

        :param instance: The :class:`Instance` (or its ID) of the database
                         instance to promote.
        """
        body = {'promote_to_replica_source': {}}
        self._action(instance, body)

    def eject_replica_source(self, instance):
        """Eject a replica source from its set

        :param instance: The :class:`Instance` (or its ID) of the database
                         instance to eject.
        """
        body = {'eject_replica_source': {}}
        self._action(instance, body)

    def modules(self, instance):
        """Get the list of modules for a specific instance."""
        return self._modules_get(instance)

    def module_query(self, instance):
        """Query an instance about installed modules."""
        return self._modules_get(instance, from_guest=True)

    def module_retrieve(self, instance, directory=None, prefix=None):
        """Retrieve the module data file from an instance.  This includes
        the contents of the module data file.
        """
        if directory:
            try:
                os.makedirs(directory, exist_ok=True)
            except TypeError:
                # py27
                try:
                    os.makedirs(directory)
                except OSError:
                    if not os.path.isdir(directory):
                        raise
        else:
            directory = '.'
        prefix = prefix or ''
        if prefix and not prefix.endswith('_'):
            prefix += '_'
        module_list = self._modules_get(
            instance, from_guest=True, include_contents=True)
        saved_modules = {}
        for module in module_list:
            filename = '%s%s_%s_%s.dat' % (prefix, module.name,
                                           module.datastore,
                                           module.datastore_version)
            full_filename = os.path.expanduser(
                os.path.join(directory, filename))
            with open(full_filename, 'wb') as fh:
                fh.write(utils.decode_data(module.contents))
            saved_modules[module.name] = full_filename
        return saved_modules

    def _modules_get(self, instance, from_guest=None, include_contents=None):
        url = "/instances/%s/modules" % base.getid(instance)
        query_strings = {}
        if from_guest is not None:
            query_strings["from_guest"] = from_guest
        if include_contents is not None:
            query_strings["include_contents"] = include_contents
        url = common.append_query_strings(url, **query_strings)
        resp, body = self.api.client.get(url)
        common.check_for_exceptions(resp, body, url)
        return [core_modules.Module(self, module, loaded=True)
                for module in body['modules']]

    def module_apply(self, instance, modules):
        """Apply modules to an instance."""
        url = "/instances/%s/modules" % base.getid(instance)
        body = {"modules": self._get_module_list(modules)}
        resp, body = self.api.client.post(url, body=body)
        common.check_for_exceptions(resp, body, url)
        return [core_modules.Module(self, module, loaded=True)
                for module in body['modules']]

    def _get_module_list(self, modules):
        """Build a list of module ids."""
        module_list = []
        for module in modules:
            module_info = {'id': base.getid(module)}
            module_list.append(module_info)
        return module_list

    def module_remove(self, instance, module):
        """Remove a module from an instance.
        """
        url = "/instances/%s/modules/%s" % (base.getid(instance),
                                            base.getid(module))
        resp, body = self.api.client.delete(url)
        common.check_for_exceptions(resp, body, url)

    def log_list(self, instance):
        """Get a list of all guest logs.

        :param instance: The :class:`Instance` (or its ID) of the database
                         instance to get the log for.
        :rtype: list of :class:`DatastoreLog`.
        """
        url = '/instances/%s/log' % base.getid(instance)
        resp, body = self.api.client.get(url)
        common.check_for_exceptions(resp, body, url)
        return [DatastoreLog(self, log, loaded=True) for log in body['logs']]

    def log_show(self, instance, log_name):
        return self.log_action(instance, log_name)

    def log_action(self, instance, log_name, enable=None, disable=None,
                   publish=None, discard=None):
        """Perform action on guest log.

        :param instance: The :class:`Instance` (or its ID) of the database
                         instance to get the log for.
        :param log_name: The name of <log> to publish
        :param enable: Turn on <log>
        :param disable: Turn off <log>
        :param publish: Publish log to associated container
        :param discard: Delete the associated container
        :rtype: List of :class:`DatastoreLog`.
        """
        body = {"name": log_name}
        if enable:
            body.update({'enable': int(enable)})
        if disable:
            body.update({'disable': int(disable)})
        if publish:
            body.update({'publish': int(publish)})
        if discard:
            body.update({'discard': int(discard)})

        url = "/instances/%s/log" % base.getid(instance)
        resp, body = self.api.client.post(url, body=body)
        common.check_for_exceptions(resp, body, url)
        return DatastoreLog(self, body['log'], loaded=True)

    def _get_container_info(self, instance, log_name):
        try:
            log_info = self.log_show(instance, log_name)
            container = log_info.container
            prefix = log_info.prefix
            metadata_file = log_info.metafile
            return container, prefix, metadata_file
        except swift_client.ClientException as ex:
            if ex.http_status == 404:
                raise exceptions.GuestLogNotFoundError()
            raise

    def log_generator(self, instance, log_name, lines=50, swift=None):
        """Return generator to yield the last <lines> lines of guest log.

        :param instance: The :class:`Instance` (or its ID) of the database
                         instance to get the log for.
        :param log_name: The name of <log> to publish
        :param lines: Display last <lines> lines of log (0 for all lines)
        :param swift: Connection to swift
        :rtype: generator function to yield log as chunks.
        """
        if not swift:
            swift = self._get_swift_client()

        def _log_generator(instance, log_name, lines, swift):
            try:
                container, prefix, metadata_file = self._get_container_info(
                    instance, log_name)

                head, body = swift.get_container(container, prefix=prefix)
                log_obj_to_display = []

                if lines:
                    total_lines = lines
                    partial_results = False
                    parts = sorted(body, key=lambda obj: obj['last_modified'],
                                   reverse=True)

                    for part in parts:
                        obj_hdrs = swift.head_object(container, part['name'])
                        obj_lines = int(obj_hdrs['x-object-meta-lines'])
                        log_obj_to_display.insert(0, part)
                        if obj_lines >= lines:
                            partial_results = True
                            break
                        lines -= obj_lines
                    if not partial_results:
                        lines = total_lines

                    part = log_obj_to_display.pop(0)
                    hdrs, log_obj = swift.get_object(container, part['name'])
                    log_by_lines = log_obj.decode().splitlines()
                    yield "\n".join(log_by_lines[-1 * lines:]) + "\n"
                else:
                    # Show all the logs
                    log_obj_to_display = sorted(
                        body, key=lambda obj: obj['last_modified'])

                for log_part in log_obj_to_display:
                    headers, log_obj = swift.get_object(container,
                                                        log_part['name'])
                    yield log_obj.decode()
            except swift_client.ClientException as ex:
                if ex.http_status == 404:
                    raise exceptions.GuestLogNotFoundError()
                raise

        return lambda: _log_generator(instance, log_name, lines, swift)

    def log_save(self, instance, log_name, filename=None):
        """Saves a guest log to a file.

        :param instance: The :class:`Instance` (or its ID) of the database
                         instance to get the log for.
        :param log_name: The name of <log> to publish
        :rtype: Filename to which log was saved
        """
        written_file = filename or (
            'trove-' + instance.id + '-' + log_name + ".log")
        log_gen = self.log_generator(instance, log_name, lines=0)
        with open(written_file, 'w') as f:
            for log_obj in log_gen():
                f.write(log_obj)
        return written_file


class InstanceStatus(object):

    ACTIVE = "ACTIVE"
    BLOCKED = "BLOCKED"
    BUILD = "BUILD"
    FAILED = "FAILED"
    REBOOT = "REBOOT"
    RESIZE = "RESIZE"
    SHUTDOWN = "SHUTDOWN"
    RESTART_REQUIRED = "RESTART_REQUIRED"
    PROMOTING = "PROMOTING"
    EJECTING = "EJECTING"
    LOGGING = "LOGGING"



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock

from osc_lib import exceptions
from osc_lib import utils

from troveclient import common
from troveclient.osc.v1 import databases
from troveclient.tests.osc.v1 import fakes


class TestDatabases(fakes.TestDatabasev1):
    fake_databases = fakes.FakeDatabases()

    def setUp(self):
        super(TestDatabases, self).setUp()
        self.mock_client = self.app.client_manager.database
        self.database_client = self.app.client_manager.database.databases


class TestDatabaseCreate(TestDatabases):

    def setUp(self):
        super(TestDatabaseCreate, self).setUp()
        self.cmd = databases.CreateDatabase(self.app, None)

    @mock.patch.object(utils, 'find_resource')
    def test_database_create(self, mock_find):
        args = ['instance1', 'db1']
        mock_find.return_value = args[0]
        parsed_args = self.check_parser(self.cmd, args, [])
        result = self.cmd.take_action(parsed_args)
        self.database_client.create.assert_called_with('instance1',
                                                       [{'name': 'db1'}])
        self.assertIsNone(result)

    @mock.patch.object(utils, 'find_resource')
    def test_database_create_with_optional_args(self, mock_find):
        args = ['instance2', 'db2',
                '--character_set', 'utf8',
                '--collate', 'utf8_general_ci']
        mock_find.return_value = args[0]
        parsed_args = self.check_parser(self.cmd, args, [])
        database_dict = {'name': 'db2',
                         'collate': 'utf8_general_ci',
                         'character_set': 'utf8'}
        result = self.cmd.take_action(parsed_args)
        self.database_client.create.assert_called_with('instance2',
                                                       [database_dict])
        self.assertIsNone(result)


class TestDatabaseList(TestDatabases):
    columns = databases.ListDatabases.columns
    values = ('fakedb1',)

    def setUp(self):
        super(TestDatabaseList, self).setUp()
        self.cmd = databases.ListDatabases(self.app, None)
        data = [self.fake_databases.get_databases_1()]
        self.database_client.list.return_value = common.Paginated(data)

    @mock.patch.object(utils, 'find_resource')
    def test_database_list_defaults(self, mock_find):
        args = ['my_instance']
        mock_find.return_value = args[0]
        parsed_args = self.check_parser(self.cmd, args, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.database_client.list.assert_called_once_with(args[0])
        self.assertEqual(self.columns, columns)
        self.assertEqual([tuple(self.values)], data)


class TestDatabaseDelete(TestDatabases):

    def setUp(self):
        super(TestDatabaseDelete, self).setUp()
        self.cmd = databases.DeleteDatabase(self.app, None)

    @mock.patch.object(utils, 'find_resource')
    def test_database_delete(self, mock_find):
        args = ['instance1', 'db1']
        mock_find.return_value = args[0]
        parsed_args = self.check_parser(self.cmd, args, [])
        result = self.cmd.take_action(parsed_args)
        self.database_client.delete.assert_called_with('instance1', 'db1')
        self.assertIsNone(result)

    @mock.patch.object(utils, 'find_resource')
    def test_database_delete_with_exception(self, mock_find):
        args = ['fakeinstance', 'db1']
        parsed_args = self.check_parser(self.cmd, args, [])

        mock_find.side_effect = exceptions.CommandError
        self.assertRaises(exceptions.CommandError,
                          self.cmd.take_action,
                          parsed_args)



# Copyright 2014 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import testtools
from troveclient.v1 import metadata
from unittest import mock


class TestMetadata(testtools.TestCase):
    def setUp(self):
        super(TestMetadata, self).setUp()
        self.orig__init = metadata.Metadata.__init__
        metadata.Metadata.__init__ = mock.Mock(return_value=None)
        self.metadata = metadata.Metadata()
        self.metadata.manager = mock.Mock()
        self.metadata.api = mock.Mock()
        self.metadata.api.client = mock.Mock()

        self.instance_uuid = '3fbc8d6d-3f87-41d9-a4a1-060830dc6c4c'
        self.metadata_key = 'metakey'
        self.new_metadata_key = 'newmetakey'
        self.metadata_value = {'metavalue': [1, 2, 3]}

    def tearDown(self):
        super(TestMetadata, self).tearDown()
        metadata.Metadata.__init__ = self.orig__init

    def test_list(self):
        def side_effect_func(path, config):
            return path, config

        self.metadata._get = mock.Mock(side_effect=side_effect_func)
        path, config = self.metadata.list(self.instance_uuid)
        self.assertEqual('/instances/%s/metadata' % self.instance_uuid, path)
        self.assertEqual('metadata', config)

    def test_show(self):
        def side_effect_func(path, config):
            return path, config

        self.metadata._get = mock.Mock(side_effect=side_effect_func)
        path, config = self.metadata.show(self.instance_uuid,
                                          self.metadata_key)
        self.assertEqual('/instances/%s/metadata/%s' %
                         (self.instance_uuid, self.metadata_key), path)
        self.assertEqual('metadata', config)

    def test_create(self):
        def side_effect_func(path, body, config):
            return path, body, config

        create_body = {
            'metadata': {
                'value': self.metadata_value
            }
        }

        self.metadata._create = mock.Mock(side_effect=side_effect_func)
        path, body, config = self.metadata.create(self.instance_uuid,
                                                  self.metadata_key,
                                                  self.metadata_value)
        self.assertEqual('/instances/%s/metadata/%s' %
                         (self.instance_uuid, self.metadata_key), path)
        self.assertEqual(create_body, body)
        self.assertEqual('metadata', config)

    def test_edit(self):
        def side_effect_func(path, body):
            return path, body

        edit_body = {
            'metadata': {
                'value': self.metadata_value
            }
        }

        self.metadata._edit = mock.Mock(side_effect=side_effect_func)
        path, body = self.metadata.edit(self.instance_uuid,
                                        self.metadata_key,
                                        self.metadata_value)
        self.assertEqual('/instances/%s/metadata/%s' %
                         (self.instance_uuid, self.metadata_key), path)
        self.assertEqual(edit_body, body)

    def test_update(self):
        def side_effect_func(path, body):
            return path, body

        update_body = {
            'metadata': {
                'key': self.new_metadata_key,
                'value': self.metadata_value
            }
        }

        self.metadata._update = mock.Mock(side_effect=side_effect_func)
        path, body = self.metadata.update(self.instance_uuid,
                                          self.metadata_key,
                                          self.new_metadata_key,
                                          self.metadata_value)
        self.assertEqual('/instances/%s/metadata/%s' %
                         (self.instance_uuid, self.metadata_key), path)
        self.assertEqual(update_body, body)

    def test_delete(self):
        def side_effect_func(path):
            return path

        self.metadata._delete = mock.Mock(side_effect=side_effect_func)
        path = self.metadata.delete(self.instance_uuid, self.metadata_key)
        self.assertEqual('/instances/%s/metadata/%s' %
                         (self.instance_uuid, self.metadata_key), path)

    def test_parse_value_valid_json_in(self):
        value = {'one': [2, 3, 4]}
        ser_value = json.dumps(value)
        new_value = self.metadata._parse_value(ser_value)
        self.assertEqual(value, new_value)

    def test_parse_value_string_in(self):
        value = 'this is a string'
        new_value = self.metadata._parse_value(value)
        self.assertEqual(value, new_value)

    def test_parse_value_dict_in(self):
        value = {'one': [2, 3, 4]}
        new_value = self.metadata._parse_value(value)
        self.assertEqual(value, new_value)

    def test_parse_value_list_in(self):
        value = [2, 3, 4]
        new_value = self.metadata._parse_value(value)
        self.assertEqual(value, new_value)

    def test_parse_value_tuple_in(self):
        value = (2, 3, 4)
        new_value = self.metadata._parse_value(value)
        self.assertEqual(value, new_value)

    def test_parse_value_float_in(self):
        value = 1.32
        new_value = self.metadata._parse_value(value)
        self.assertEqual(value, new_value)

    def test_parse_value_int_in(self):
        value = 1
        new_value = self.metadata._parse_value(value)
        self.assertEqual(value, new_value)

    def test_parse_value_invalid_json_in(self):
        # NOTE(imsplitbit): it's worth mentioning here and in the code that
        # if you give _parse_value invalid json you get the string passed back
        # to you.
        value = "{'one': [2, 3, 4]}"
        new_value = self.metadata._parse_value(value)
        self.assertEqual(value, new_value)



# Copyright 2016 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from troveclient import base
from troveclient import common
from troveclient import utils


class Module(base.Resource):

    ALL_KEYWORD = 'all'

    def __repr__(self):
        return "<Module: %s>" % self.name

    def __hash__(self):
        return hash(repr(self))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False


class Modules(base.ManagerWithFind):
    """Manage :class:`Module` resources."""
    resource_class = Module

    def create(self, name, module_type, contents, description=None,
               all_tenants=None, datastore=None,
               datastore_version=None, auto_apply=None,
               visible=None, live_update=None,
               priority_apply=None, apply_order=None,
               full_access=None):
        """Create a new module."""

        contents = utils.encode_data(contents)
        body = {"module": {
            "name": name,
            "module_type": module_type,
            "contents": contents,
        }}
        if description is not None:
            body["module"]["description"] = description
        datastore_obj = {}
        if datastore:
            datastore_obj["type"] = datastore
        if datastore_version:
            datastore_obj["version"] = datastore_version
        if datastore_obj:
            body["module"]["datastore"] = datastore_obj
        if all_tenants is not None:
            body["module"]["all_tenants"] = int(all_tenants)
        if auto_apply is not None:
            body["module"]["auto_apply"] = int(auto_apply)
        if visible is not None:
            body["module"]["visible"] = int(visible)
        if live_update is not None:
            body["module"]["live_update"] = int(live_update)
        if priority_apply is not None:
            body["module"]["priority_apply"] = int(priority_apply)
        if apply_order is not None:
            body["module"]["apply_order"] = apply_order
        if full_access is not None:
            body["module"]["full_access"] = int(full_access)

        return self._create("/modules", body, "module")

    def update(self, module, name=None, module_type=None,
               contents=None, description=None,
               all_tenants=None, datastore=None,
               datastore_version=None, auto_apply=None,
               visible=None, live_update=None,
               all_datastores=None, all_datastore_versions=None,
               priority_apply=None, apply_order=None,
               full_access=None):
        """Update an existing module. Passing in
        datastore=None or datastore_version=None has the effect of
        making it available for all datastores/versions.
        """
        body = {
            "module": {
            }
        }
        if name is not None:
            body["module"]["name"] = name
        if module_type is not None:
            body["module"]["type"] = module_type
        if contents is not None:
            contents = utils.encode_data(contents)
            body["module"]["contents"] = contents
        if description is not None:
            body["module"]["description"] = description
        datastore_obj = {}
        if datastore:
            datastore_obj["type"] = datastore
        if datastore_version:
            datastore_obj["version"] = datastore_version
        if datastore_obj:
            body["module"]["datastore"] = datastore_obj
        if all_datastores:
            body["module"]["all_datastores"] = int(all_datastores)
        if all_datastore_versions:
            body["module"]["all_datastore_versions"] = int(
                all_datastore_versions)
        if all_tenants is not None:
            body["module"]["all_tenants"] = int(all_tenants)
        if auto_apply is not None:
            body["module"]["auto_apply"] = int(auto_apply)
        if visible is not None:
            body["module"]["visible"] = int(visible)
        if live_update is not None:
            body["module"]["live_update"] = int(live_update)
        if priority_apply is not None:
            body["module"]["priority_apply"] = int(priority_apply)
        if apply_order is not None:
            body["module"]["apply_order"] = apply_order
        if full_access is not None:
            body["module"]["full_access"] = int(full_access)

        url = "/modules/%s" % base.getid(module)
        resp, body = self.api.client.put(url, body=body)
        common.check_for_exceptions(resp, body, url)
        return Module(self, body['module'], loaded=True)

    def list(self, limit=None, marker=None, datastore=None):
        """Get a list of all modules."""
        query_strings = None
        if datastore:
            query_strings = {"datastore": base.getid(datastore)}
        return self._paginated(
            "/modules", "modules", limit, marker, query_strings=query_strings)

    def get(self, module):
        """Get a specific module."""
        return self._get(
            "/modules/%s" % base.getid(module), "module")

    def delete(self, module):
        """Delete the specified module."""
        url = "/modules/%s" % base.getid(module)
        resp, body = self.api.client.delete(url)
        common.check_for_exceptions(resp, body, url)

    def instances(self, module, limit=None, marker=None,
                  include_clustered=False, count_only=False):
        """Get a list of all instances this module has been applied to."""
        url = "/modules/%s/instances" % base.getid(module)
        query_strings = {}
        if include_clustered:
            query_strings['include_clustered'] = include_clustered
        if count_only:
            query_strings['count_only'] = count_only
        return self._paginated(url, "instances", limit, marker,
                               query_strings=query_strings)

    def reapply(self, module, md5=None, include_clustered=None,
                batch_size=None, delay=None, force=None):
        """Reapplies the specified module."""
        url = "/modules/%s/instances" % base.getid(module)
        body = {
            "reapply": {
            }
        }
        if md5:
            body["reapply"]["md5"] = md5
        if include_clustered is not None:
            body["reapply"]["include_clustered"] = int(include_clustered)
        if batch_size is not None:
            body["reapply"]["batch_size"] = batch_size
        if delay is not None:
            body["reapply"]["batch_delay"] = delay
        if force is not None:
            body["reapply"]["force"] = int(force)
        resp, body = self.api.client.put(url, body=body)
        common.check_for_exceptions(resp, body, url)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock

from osc_lib import exceptions
from osc_lib import utils

from troveclient import common
from troveclient.osc.v1 import database_clusters
from troveclient.tests.osc.v1 import fakes


class TestClusters(fakes.TestDatabasev1):
    fake_clusters = fakes.FakeClusters()

    def setUp(self):
        super(TestClusters, self).setUp()
        self.mock_client = self.app.client_manager.database
        self.cluster_client = self.app.client_manager.database.clusters
        self.instance_client = self.app.client_manager.database.instances


class TestClusterList(TestClusters):

    defaults = {
        'limit': None,
        'marker': None
    }

    columns = database_clusters.ListDatabaseClusters.columns
    values = ('cls-1234', 'test-clstr', 'vertica', '7.1', 'NONE')

    def setUp(self):
        super(TestClusterList, self).setUp()
        self.cmd = database_clusters.ListDatabaseClusters(self.app, None)
        data = [self.fake_clusters.get_clusters_cls_1234()]
        self.cluster_client.list.return_value = common.Paginated(data)

    def test_cluster_list_defaults(self):
        parsed_args = self.check_parser(self.cmd, [], [])
        columns, data = self.cmd.take_action(parsed_args)
        self.cluster_client.list.assert_called_once_with(**self.defaults)
        self.assertEqual(self.columns, columns)
        self.assertEqual([self.values], data)


class TestClusterShow(TestClusters):

    values = ('2015-05-02T10:37:04', 'vertica', '7.1', 'cls-1234', 2,
              'test-clstr', 'No tasks for the cluster.', 'NONE',
              '2015-05-02T11:06:19')

    def setUp(self):
        super(TestClusterShow, self).setUp()
        self.cmd = database_clusters.ShowDatabaseCluster(self.app, None)
        self.data = self.fake_clusters.get_clusters_cls_1234()
        self.cluster_client.get.return_value = self.data
        self.columns = (
            'created',
            'datastore',
            'datastore_version',
            'id',
            'instance_count',
            'name',
            'task_description',
            'task_name',
            'updated',
        )

    def test_show(self):
        args = ['cls-1234']
        parsed_args = self.check_parser(self.cmd, args, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.values, data)


class TestDatabaseClusterDelete(TestClusters):

    def setUp(self):
        super(TestDatabaseClusterDelete, self).setUp()
        self.cmd = database_clusters.DeleteDatabaseCluster(self.app, None)

    @mock.patch.object(utils, 'find_resource')
    def test_cluster_delete(self, mock_find):
        args = ['cluster1']
        mock_find.return_value = args[0]
        parsed_args = self.check_parser(self.cmd, args, [])
        result = self.cmd.take_action(parsed_args)
        self.cluster_client.delete.assert_called_with('cluster1')
        self.assertIsNone(result)

    @mock.patch.object(utils, 'find_resource')
    def test_cluster_delete_with_exception(self, mock_find):
        args = ['fakecluster']
        parsed_args = self.check_parser(self.cmd, args, [])

        mock_find.side_effect = exceptions.CommandError
        self.assertRaises(exceptions.CommandError,
                          self.cmd.take_action,
                          parsed_args)


class TestDatabaseClusterCreate(TestClusters):

    values = ('2015-05-02T10:37:04', 'vertica', '7.1', 'cls-1234',
              2, 'test-clstr', 'No tasks for the cluster.',
              'NONE', '2015-05-02T11:06:19')
    columns = (
        'created',
        'datastore',
        'datastore_version',
        'id',
        'instance_count',
        'name',
        'task_description',
        'task_name',
        'updated',
    )

    def setUp(self):
        super(TestDatabaseClusterCreate, self).setUp()
        self.cmd = database_clusters.CreateDatabaseCluster(self.app, None)
        self.data = self.fake_clusters.get_clusters_cls_1234()
        self.cluster_client.create.return_value = self.data

    @mock.patch.object(database_clusters, '_parse_instance_options')
    def test_cluster_create(self, mock_parse_instance_opts):
        instances = ['flavor="02",volume=2',
                     'flavor="03",volume=3']
        parsed_instances = [{'flavor': '02', 'volume': 2},
                            {'flavor': '03', 'volume': 3}]
        extended_properties = "foo_properties=foo_value"
        parsed_extended_properties = {'foo_properties': 'foo_value'}
        mock_parse_instance_opts.return_value = parsed_instances
        args = ['test-name', 'vertica', '7.1',
                '--instance', instances[0],
                '--instance', instances[1],
                '--extended-properties', extended_properties,
                '--configuration', 'config01']
        verifylist = [
            ('name', 'test-name'),
            ('datastore', 'vertica'),
            ('datastore_version', '7.1'),
            ('instances', instances),
            ('extended_properties', extended_properties),
            ('configuration', 'config01'),
        ]
        parsed_args = self.check_parser(self.cmd, args, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.cluster_client.create.assert_called_with(
            parsed_args.name, parsed_args.datastore,
            parsed_args.datastore_version,
            instances=parsed_instances,
            locality=parsed_args.locality,
            extended_properties=parsed_extended_properties,
            configuration=parsed_args.configuration)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.values, data)


class TestDatabaseClusterResetStatus(TestClusters):

    def setUp(self):
        super(TestDatabaseClusterResetStatus, self).setUp()
        self.cmd = database_clusters.ResetDatabaseClusterStatus(self.app, None)

    @mock.patch.object(utils, 'find_resource')
    def test_cluster_reset_status(self, mock_find):
        args = ['cluster1']
        mock_find.return_value = args[0]
        parsed_args = self.check_parser(self.cmd, args, [])
        result = self.cmd.take_action(parsed_args)
        self.cluster_client.reset_status.assert_called_with('cluster1')
        self.assertIsNone(result)


class TestClusterListInstances(TestClusters):

    columns = (database_clusters
               .ListDatabaseClusterInstances.columns)
    values = [('member-1', 'test-clstr-member-1', '02', 2, 'ACTIVE'),
              ('member-2', 'test-clstr-member-2', '2', 2, 'ACTIVE')]

    def setUp(self):
        super(TestClusterListInstances, self).setUp()
        self.cmd = (database_clusters
                    .ListDatabaseClusterInstances(self.app, None))
        self.data = self.fake_clusters.get_clusters_cls_1234()
        self.cluster_client.get.return_value = self.data

    def test_cluster_list_instances(self):
        args = ['cls-1234']
        parsed_args = self.check_parser(self.cmd, args, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.values, data)


class TestDatabaseClusterUpgrade(TestClusters):

    def setUp(self):
        super(TestDatabaseClusterUpgrade, self).setUp()
        self.cmd = database_clusters.UpgradeDatabaseCluster(self.app, None)

    @mock.patch.object(utils, 'find_resource')
    def test_cluster_upgrade(self, mock_find):
        args = ['cluster1', 'datastore_version1']
        mock_find.return_value = args[0]
        parsed_args = self.check_parser(self.cmd, args, [])
        result = self.cmd.take_action(parsed_args)
        self.cluster_client.upgrade.assert_called_with('cluster1',
                                                       'datastore_version1')
        self.assertIsNone(result)


class TestDatabaseClusterForceDelete(TestClusters):

    def setUp(self):
        super(TestDatabaseClusterForceDelete, self).setUp()
        self.cmd = database_clusters.ForceDeleteDatabaseCluster(self.app, None)

    @mock.patch.object(utils, 'find_resource')
    def test_cluster_force_delete(self, mock_find):
        args = ['cluster1']
        mock_find.return_value = args[0]
        parsed_args = self.check_parser(self.cmd, args, [])
        result = self.cmd.take_action(parsed_args)
        self.cluster_client.reset_status.assert_called_with('cluster1')
        self.cluster_client.delete.assert_called_with('cluster1')
        self.assertIsNone(result)

    @mock.patch.object(utils, 'find_resource')
    def test_cluster_force_delete_with_exception(self, mock_find):
        args = ['fakecluster']
        parsed_args = self.check_parser(self.cmd, args, [])
        mock_find.return_value = args[0]

        self.cluster_client.delete.side_effect = exceptions.CommandError
        self.assertRaises(exceptions.CommandError,
                          self.cmd.take_action,
                          parsed_args)


class TestDatabaseClusterGrow(TestClusters):

    def setUp(self):
        super(TestDatabaseClusterGrow, self).setUp()
        self.cmd = database_clusters.GrowDatabaseCluster(self.app, None)

    @mock.patch.object(utils, 'find_resource')
    @mock.patch.object(database_clusters, '_parse_instance_options')
    def test_cluster_grow(self, mock_parse_instance_opts,
                          mock_find_resource):
        args = ['test-clstr',
                '--instance',
                'name=test-clstr-member-3,flavor=3']
        parsed_instance_opts = [
            {'name': 'test-clstr-member-3',
             'flavor': 3}
        ]
        mock_parse_instance_opts.return_value = parsed_instance_opts
        mock_find_resource.return_value = args[0]

        parsed_args = self.check_parser(self.cmd, args, [])
        result = self.cmd.take_action(parsed_args)
        self.cluster_client.grow.assert_called_with(
            'test-clstr',
            instances=parsed_instance_opts)
        self.assertIsNone(result)


class TestDatabaseClusterShrink(TestClusters):

    def setUp(self):
        super(TestDatabaseClusterShrink, self).setUp()
        self.cmd = database_clusters.ShrinkDatabaseCluster(self.app, None)
        self.cluster_member = self.fake_clusters.get_clusters_member_2()

    @mock.patch.object(utils, 'find_resource')
    def test_cluster_grow(self, mock_find_resource):
        args = ['test-clstr', 'test-clstr-member-2']
        mock_find_resource.side_effect = [
            args[0], self.cluster_member]

        parsed_args = self.check_parser(self.cmd, args, [])
        result = self.cmd.take_action(parsed_args)
        self.cluster_client.shrink.assert_called_with('test-clstr',
                                                      [{'id': 'member-2'}])
        self.assertIsNone(result)


class TestClusterListModules(TestClusters):

    columns = (database_clusters
               .ListDatabaseClusterModules.columns)
    values = [('test-clstr-member-1', 'mymod1', 'ping', 'md5-1',
               '2018-04-17 05:34:02.84', '2018-04-17 05:34:02.84'),
              ('test-clstr-member-2', 'mymod2', 'ping', 'md5-2',
               '2018-04-17 05:34:02.84', '2018-04-17 05:34:02.84')]

    def setUp(self):
        super(TestClusterListModules, self).setUp()
        self.cmd = (database_clusters
                    .ListDatabaseClusterModules(self.app, None))
        self.data = self.fake_clusters.get_clusters_cls_1234()
        self.instance_client.modules.side_effect = (
            self.fake_clusters.cluster_instance_modules())

    @mock.patch.object(utils, 'find_resource')
    def test_cluster_list_modules(self, mock_find_resource):
        mock_find_resource.return_value = self.data
        args = ['cls-1234']
        parsed_args = self.check_parser(self.cmd, args, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.values, data)



# Copyright 2011 OpenStack Foundation
# Copyright 2013 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from six.moves.urllib import parse

from troveclient.apiclient import exceptions


def check_for_exceptions(resp, body, url):
    if resp.status_code in (400, 422, 500):
        raise exceptions.from_response(resp, body, url)


def append_query_strings(url, **query_strings):
    if not query_strings:
        return url
    query = '&'.join('{0}={1}'.format(key, val)
                     for key, val in query_strings.items() if val is not None)
    return url + ('?' + query if query else "")


def quote_user_host(user, host):
    quoted = ''
    if host:
        quoted = parse.quote("%s@%s" % (user, host))
    else:
        quoted = parse.quote("%s" % user)
    return quoted.replace('.', '%2e')


class Paginated(list):

    def __init__(self, items=None, next_marker=None, links=None):
        items = items or []
        links = links or []
        super(Paginated, self).__init__(items)
        self.next = next_marker
        self.links = links



# Copyright (c) 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import httplib2
import logging
import os
import sys
import time

try:
    import json
except ImportError:
    import simplejson as json


from troveclient.compat import auth
from troveclient.compat import exceptions


LOG = logging.getLogger(__name__)
RDC_PP = os.environ.get("RDC_PP", "False") == "True"


expected_errors = (400, 401, 403, 404, 408, 409, 413, 422, 500, 501)


def log_to_streamhandler(stream=None):
    stream = stream or sys.stderr
    ch = logging.StreamHandler(stream)
    LOG.setLevel(logging.DEBUG)
    LOG.addHandler(ch)


if 'REDDWARFCLIENT_DEBUG' in os.environ and os.environ['REDDWARFCLIENT_DEBUG']:
    log_to_streamhandler()


class TroveHTTPClient(httplib2.Http):

    USER_AGENT = 'python-troveclient'

    def __init__(self, user, password, tenant, auth_url, service_name,
                 service_url=None,
                 auth_strategy=None, insecure=False,
                 timeout=None, proxy_tenant_id=None,
                 proxy_token=None, region_name=None,
                 endpoint_type='publicURL', service_type=None,
                 timings=False):

        super(TroveHTTPClient, self).__init__(timeout=timeout)

        self.username = user
        self.password = password
        self.tenant = tenant
        if auth_url:
            self.auth_url = auth_url.rstrip('/')
        else:
            self.auth_url = None
        self.region_name = region_name
        self.endpoint_type = endpoint_type
        self.service_url = service_url
        self.service_type = service_type
        self.service_name = service_name
        self.timings = timings

        self.times = []  # [("item", starttime, endtime), ...]

        self.auth_token = None
        self.proxy_token = proxy_token
        self.proxy_tenant_id = proxy_tenant_id

        # httplib2 overrides
        self.force_exception_to_status_code = True
        self.disable_ssl_certificate_validation = insecure

        auth_cls = auth.get_authenticator_cls(auth_strategy)

        self.authenticator = auth_cls(self, auth_strategy,
                                      self.auth_url, self.username,
                                      self.password, self.tenant,
                                      region=region_name,
                                      service_type=service_type,
                                      service_name=service_name,
                                      service_url=service_url)
        if hasattr(self.authenticator, 'auth'):
            self.auth = self.authenticator.auth

    def get_timings(self):
        return self.times

    def http_log(self, args, kwargs, resp, body):
        if not RDC_PP:
            self.simple_log(args, kwargs, resp, body)
        else:
            self.pretty_log(args, kwargs, resp, body)

    def simple_log(self, args, kwargs, resp, body):
        if not LOG.isEnabledFor(logging.DEBUG):
            return

        string_parts = ['curl -i']
        for element in args:
            if element in ('GET', 'POST'):
                string_parts.append(' -X %s' % element)
            else:
                string_parts.append(' %s' % element)

        for element in kwargs['headers']:
            header = ' -H "%s: %s"' % (element, kwargs['headers'][element])
            string_parts.append(header)

        LOG.debug("REQ: %s\n", "".join(string_parts))
        if 'body' in kwargs:
            LOG.debug("REQ BODY: %s\n", kwargs['body'])
        LOG.debug("RESP:%s %s\n", resp, body)

    def pretty_log(self, args, kwargs, resp, body):
        if not LOG.isEnabledFor(logging.DEBUG):
            return

        string_parts = ['curl -i']
        for element in args:
            if element in ('GET', 'POST'):
                string_parts.append(' -X %s' % element)
            else:
                string_parts.append(' %s' % element)

        for element in kwargs['headers']:
            header = ' -H "%s: %s"' % (element, kwargs['headers'][element])
            string_parts.append(header)

        curl_cmd = "".join(string_parts)
        LOG.debug("REQUEST:")
        if 'body' in kwargs:
            LOG.debug("%s -d '%s'", curl_cmd, kwargs['body'])
            try:
                req_body = json.dumps(json.loads(kwargs['body']),
                                      sort_keys=True, indent=4)
            except Exception:
                req_body = kwargs['body']
            LOG.debug("BODY: %s\n", req_body)
        else:
            LOG.debug(curl_cmd)

        try:
            resp_body = json.dumps(json.loads(body), sort_keys=True, indent=4)
        except Exception:
            resp_body = body
        LOG.debug("RESPONSE HEADERS: %s", resp)
        LOG.debug("RESPONSE BODY   : %s", resp_body)

    def request(self, *args, **kwargs):
        kwargs.setdefault('headers', kwargs.get('headers', {}))
        kwargs['headers']['User-Agent'] = self.USER_AGENT
        self.morph_request(kwargs)

        resp, body = super(TroveHTTPClient, self).request(*args, **kwargs)
        # compat between requests and httplib2
        resp.status_code = resp.status

        # Save this in case anyone wants it.
        self.last_response = (resp, body)
        self.http_log(args, kwargs, resp, body)

        if body:
            try:
                body = self.morph_response_body(body)
            except exceptions.ResponseFormatError:
                # Acceptable only if the response status is an error code.
                # Otherwise its the API or client misbehaving.
                self.raise_error_from_status(resp, None)
                raise  # Not accepted!
        else:
            body = None

        if resp.status in expected_errors:
            raise exceptions.from_response(resp, body)

        return resp, body

    def raise_error_from_status(self, resp, body):
        if resp.status in expected_errors:
            raise exceptions.from_response(resp, body)

    def morph_request(self, kwargs):
        kwargs['headers']['Accept'] = 'application/json'
        kwargs['headers']['Content-Type'] = 'application/json'
        if 'body' in kwargs:
            kwargs['body'] = json.dumps(kwargs['body'])

    def morph_response_body(self, raw_body):
        try:
            return json.loads(raw_body.decode())
        except ValueError:
            raise exceptions.ResponseFormatError()

    def _time_request(self, url, method, **kwargs):
        start_time = time.time()
        resp, body = self.request(url, method, **kwargs)
        self.times.append(("%s %s" % (method, url),
                           start_time, time.time()))
        return resp, body

    def _cs_request(self, url, method, **kwargs):
        def request():
            kwargs.setdefault('headers', {})['X-Auth-Token'] = self.auth_token
            if self.tenant:
                kwargs['headers']['X-Auth-Project-Id'] = self.tenant

            resp, body = self._time_request(self.service_url + url, method,
                                            **kwargs)
            return resp, body

        if not self.auth_token or not self.service_url:
            self.authenticate()

        # Perform the request once. If we get a 401 back then it
        # might be because the auth token expired, so try to
        # re-authenticate and try again. If it still fails, bail.
        try:
            return request()
        except exceptions.Unauthorized:
            self.authenticate()
            return request()

    def get(self, url, **kwargs):
        return self._cs_request(url, 'GET', **kwargs)

    def patch(self, url, **kwargs):
        return self._cs_request(url, 'PATCH', **kwargs)

    def post(self, url, **kwargs):
        return self._cs_request(url, 'POST', **kwargs)

    def put(self, url, **kwargs):
        return self._cs_request(url, 'PUT', **kwargs)

    def delete(self, url, **kwargs):
        return self._cs_request(url, 'DELETE', **kwargs)

    def authenticate(self):
        """Auths the client and gets a token. May optionally set a service url.

        The client will get auth errors until the authentication step
        occurs. Additionally, if a service_url was not explicitly given in
        the clients __init__ method, one will be obtained from the auth
        service.

        """
        catalog = self.authenticator.authenticate()
        if self.service_url:
            possible_service_url = None
        else:
            if self.endpoint_type == "publicURL":
                possible_service_url = catalog.get_public_url()
            elif self.endpoint_type == "adminURL":
                possible_service_url = catalog.get_management_url()
        self.authenticate_with_token(catalog.get_token(), possible_service_url)

    def authenticate_with_token(self, token, service_url=None):
        self.auth_token = token
        if not self.service_url:
            if not service_url:
                raise exceptions.ServiceUrlNotGiven()
            else:
                self.service_url = service_url


class Dbaas(object):
    """Top-level object to access the Rackspace Database as a Service API.

    Create an instance with your creds::

        >> red = Dbaas(USERNAME, API_KEY, TENANT, AUTH_URL, SERVICE_NAME, \
                        SERVICE_URL)

    Then call methods on its managers::

        >> red.instances.list()
        ...
        >> red.flavors.list()
        ...

    &c.
    """

    def __init__(self, username, api_key, tenant=None, auth_url=None,
                 service_type='database', service_name=None,
                 service_url=None, insecure=False, auth_strategy='keystone',
                 region_name=None, client_cls=TroveHTTPClient):

        from troveclient.compat import versions
        from troveclient.v1 import accounts
        from troveclient.v1 import backups
        from troveclient.v1 import clusters
        from troveclient.v1 import configurations
        from troveclient.v1 import databases
        from troveclient.v1 import datastores
        from troveclient.v1 import diagnostics
        from troveclient.v1 import hosts
        from troveclient.v1 import instances
        from troveclient.v1 import limits
        from troveclient.v1 import management
        from troveclient.v1 import metadata
        from troveclient.v1 import modules
        from troveclient.v1 import quota
        from troveclient.v1 import root
        from troveclient.v1 import security_groups
        from troveclient.v1 import storage
        from troveclient.v1 import users

        self.client = client_cls(username, api_key, tenant, auth_url,
                                 service_type=service_type,
                                 service_name=service_name,
                                 service_url=service_url,
                                 insecure=insecure,
                                 auth_strategy=auth_strategy,
                                 region_name=region_name)
        self.versions = versions.Versions(self)
        self.databases = databases.Databases(self)
        self.instances = instances.Instances(self)
        self.limits = limits.Limits(self)
        self.users = users.Users(self)
        self.root = root.Root(self)
        self.hosts = hosts.Hosts(self)
        self.quota = quota.Quotas(self)
        self.backups = backups.Backups(self)
        self.clusters = clusters.Clusters(self)
        self.security_groups = security_groups.SecurityGroups(self)
        self.security_group_rules = security_groups.SecurityGroupRules(self)
        self.datastores = datastores.Datastores(self)
        self.datastore_versions = datastores.DatastoreVersions(self)
        self.datastore_version_members = (datastores.
                                          DatastoreVersionMembers(self))
        self.storage = storage.StorageInfo(self)
        self.management = management.Management(self)
        self.mgmt_cluster = management.MgmtClusters(self)
        self.accounts = accounts.Accounts(self)
        self.diagnostics = diagnostics.DiagnosticsInterrogator(self)
        self.hwinfo = diagnostics.HwInfoInterrogator(self)
        self.configurations = configurations.Configurations(self)
        config_parameters = configurations.ConfigurationParameters(self)
        self.configuration_parameters = config_parameters
        self.metadata = metadata.Metadata(self)
        self.modules = modules.Modules(self)
        self.mgmt_configs = management.MgmtConfigurationParameters(self)
        self.mgmt_datastore_versions = management.MgmtDatastoreVersions(self)

        class Mgmt(object):
            def __init__(self, dbaas):
                self.instances = dbaas.management
                self.hosts = dbaas.hosts
                self.accounts = dbaas.accounts
                self.storage = dbaas.storage
                self.datastore_version = dbaas.mgmt_datastore_versions

        self.mgmt = Mgmt(self)

    def set_management_url(self, url):
        self.client.management_url = url

    def get_timings(self):
        return self.client.get_timings()

    def authenticate(self):
        """Authenticate against the server.

        This is called to perform an authentication to retrieve a token.

        Returns on success; raises :exc:`exceptions.Unauthorized` if the
        credentials are wrong.
        """
        self.client.authenticate()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Database v1 Databases action implementations"""

from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils

from troveclient.i18n import _


class CreateDatabase(command.Command):

    _description = _("Creates a database on an instance.")

    def get_parser(self, prog_name):
        parser = super(CreateDatabase, self).get_parser(prog_name)
        parser.add_argument(
            "instance",
            metavar="<instance>",
            help=_("ID or name of the instance."),
        )
        parser.add_argument(
            "name",
            metavar="<name>",
            help=_("Name of the database."),
        )
        parser.add_argument(
            "--character_set",
            metavar="<character_set>",
            help=_("Optional character set for database."),
        )
        parser.add_argument(
            "--collate",
            metavar="<collate>",
            help=_("Optional collation type for database."),
        )
        return parser

    def take_action(self, parsed_args):
        manager = self.app.client_manager.database
        databases = manager.databases
        instance = utils.find_resource(manager.instances, parsed_args.instance)
        database_dict = {'name': parsed_args.name}
        if parsed_args.collate:
            database_dict['collate'] = parsed_args.collate
        if parsed_args.character_set:
            database_dict['character_set'] = parsed_args.character_set
        databases.create(instance, [database_dict])


class ListDatabases(command.Lister):

    _description = _("Get a list of all Databases from the instance.")
    columns = ['Name']

    def get_parser(self, prog_name):
        parser = super(ListDatabases, self).get_parser(prog_name)
        parser.add_argument(
            dest='instance',
            metavar='<instance>',
            help=_('ID or name of the instance.')
        )
        return parser

    def take_action(self, parsed_args):
        manager = self.app.client_manager.database
        databases = manager.databases
        instance = utils.find_resource(manager.instances, parsed_args.instance)
        items = databases.list(instance)
        dbs = items
        while items.next:
            items = databases.list(instance, marker=items.next)
            dbs += items
        dbs = [utils.get_item_properties(db, self.columns) for db in dbs]
        return self.columns, dbs


class DeleteDatabase(command.Command):

    _description = _("Deletes a database from an instance.")

    def get_parser(self, prog_name):
        parser = super(DeleteDatabase, self).get_parser(prog_name)
        parser.add_argument(
            dest='instance',
            metavar='<instance>',
            help=_('ID or name of the instance.')
        )
        parser.add_argument(
            dest='database',
            metavar='<database>',
            help=_('Name of the database.')
        )
        return parser

    def take_action(self, parsed_args):
        manager = self.app.client_manager.database
        databases = manager.databases
        try:
            instance = utils.find_resource(manager.instances,
                                           parsed_args.instance)
            databases.delete(instance, parsed_args.database)
        except Exception as e:
            msg = (_("Failed to delete database %(database)s: %(e)s")
                   % {'database': parsed_args.database, 'e': e})
            raise exceptions.CommandError(msg)



# Copyright 2011 OpenStack Foundation
# Copyright 2013 Rackspace Hosting
# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools
from unittest import mock

from troveclient.v1 import limits


class LimitsTest(testtools.TestCase):
    """This class tests the calling code for the Limits API."""

    def setUp(self):
        super(LimitsTest, self).setUp()
        self.limits = limits.Limits(mock.Mock())
        self.limits.api.client = mock.Mock()

    def tearDown(self):
        super(LimitsTest, self).tearDown()

    def test_list(self):
        resp = mock.Mock()
        resp.status_code = 200
        body = {"limits":
                [
                    {'maxTotalInstances': 55,
                     'verb': 'ABSOLUTE',
                     'maxTotalVolumes': 100},
                    {'regex': '.*',
                     'nextAvailable': '2011-07-21T18:17:06Z',
                     'uri': '*',
                     'value': 10,
                     'verb': 'POST',
                     'remaining': 2, 'unit': 'MINUTE'},
                    {'regex': '.*',
                     'nextAvailable': '2011-07-21T18:17:06Z',
                     'uri': '*',
                     'value': 10,
                     'verb': 'PUT',
                     'remaining': 2,
                     'unit': 'MINUTE'},
                    {'regex': '.*',
                     'nextAvailable': '2011-07-21T18:17:06Z',
                     'uri': '*',
                     'value': 10,
                     'verb': 'DELETE',
                     'remaining': 2,
                     'unit': 'MINUTE'},
                    {'regex': '.*',
                     'nextAvailable': '2011-07-21T18:17:06Z',
                     'uri': '*',
                     'value': 10,
                     'verb': 'GET',
                     'remaining': 2, 'unit': 'MINUTE'}]}
        response = (resp, body)

        mock_get = mock.Mock(return_value=response)
        self.limits.api.client.get = mock_get
        self.assertIsNotNone(self.limits.list())
        mock_get.assert_called_once_with("/limits")

    def test_list_errors(self):
        status_list = [400, 401, 403, 404, 408, 409, 413, 500, 501]
        for status_code in status_list:
            self._check_error_response(status_code)

    def _check_error_response(self, status_code):
        RESPONSE_KEY = "limits"

        resp = mock.Mock()
        resp.status_code = status_code
        body = {RESPONSE_KEY: {
            'absolute': {},
            'rate': [
                {'limit': []}]}}
        response = (resp, body)

        mock_get = mock.Mock(return_value=response)
        self.limits.api.client.get = mock_get
        self.assertRaises(Exception, self.limits.list)



# Copyright 2011 OpenStack Foundation
# Copyright 2013 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from troveclient import base
from troveclient import common


class Host(base.Resource):
    """A Hosts is an opaque instance used to store Host instances."""
    def __repr__(self):
        return "<Host: %s>" % self.name


class Hosts(base.ManagerWithFind):
    """Manage :class:`Host` resources."""
    resource_class = Host

    def _list(self, url, response_key):
        resp, body = self.api.client.get(url)
        if not body:
            raise Exception("Call to " + url + " did not return a body.")
        return [self.resource_class(self, res) for res in body[response_key]]

    def _action(self, host_id, body):
        """Perform a host "action" -- update."""
        url = "/mgmt/hosts/%s/instances/action" % host_id
        resp, body = self.api.client.post(url, body=body)
        common.check_for_exceptions(resp, body, url)

    def update_all(self, host_id):
        """Update all instances on a host."""
        body = {'update': ''}
        self._action(host_id, body)

    def index(self):
        """Get a list of all hosts.

        :rtype: list of :class:`Hosts`.
        """
        return self._list("/mgmt/hosts", "hosts")

    def get(self, host):
        """Get a specific host.

        :rtype: :class:`host`
        """
        return self._get("/mgmt/hosts/%s" % self._get_host_name(host), "host")

    @staticmethod
    def _get_host_name(host):
        try:
            if host.name:
                return host.name
        except AttributeError:
            return host

    # Appease the abc gods
    def list(self):
        pass



# Copyright 2011 OpenStack Foundation
# Copyright 2013 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from troveclient import base
from troveclient import common


class Database(base.Resource):
    """Wikipedia definition for database.

    "A database is a system intended to organize, store, and retrieve large
    amounts of data easily."
    """
    def __repr__(self):
        return "<Database: %s>" % self.name


class Databases(base.ManagerWithFind):
    """Manage :class:`Databases` resources."""
    resource_class = Database

    def create(self, instance, databases):
        """Create new databases within the specified instance."""
        body = {"databases": databases}
        url = "/instances/%s/databases" % base.getid(instance)
        resp, body = self.api.client.post(url, body=body)
        common.check_for_exceptions(resp, body, url)

    def delete(self, instance, dbname):
        """Delete an existing database in the specified instance."""
        url = "/instances/%s/databases/%s" % (base.getid(instance), dbname)
        resp, body = self.api.client.delete(url)
        common.check_for_exceptions(resp, body, url)

    def list(self, instance, limit=None, marker=None):
        """Get a list of all Databases from the instance.

        :rtype: list of :class:`Database`.
        """
        url = "/instances/%s/databases" % base.getid(instance)
        return self._paginated(url, "databases", limit, marker)

#    def get(self, instance, database):
#        """
#        Get a specific instances.
#
#        :param flavor: The ID of the :class:`Database` to get.
#        :rtype: :class:`Database`
#        """
#        assert isinstance(instance, Instance)
#        assert isinstance(database, (Database, int))
#        instance_id = base.getid(instance)
#        db_id = base.getid(database)
#        url = "/instances/%s/databases/%s" % (instance_id, db_id)
#        return self._get(url, "database")



# Copyright 2011 OpenStack Foundation
# Copyright 2013 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from troveclient import base


class Device(base.Resource):
    """Storage is an opaque instance used to hold storage information."""
    def __repr__(self):
        return "<Device: %s>" % self.name


class StorageInfo(base.ManagerWithFind):
    """Manage :class:`Storage` resources."""
    resource_class = Device

    def _list(self, url, response_key):
        resp, body = self.api.client.get(url)
        if not body:
            raise Exception("Call to " + url + " did not return a body.")
        return [self.resource_class(self, res) for res in body[response_key]]

    def index(self):
        """Get a list of all storages.

        :rtype: list of :class:`Storages`.
        """
        return self._list("/mgmt/storage", "devices")

    # Appease the abc gods
    def list(self):
        pass



# Copyright 2011 OpenStack Foundation
# Copyright 2013 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from troveclient import base
from troveclient import common
from troveclient import exceptions
from troveclient.v1 import databases


class User(base.Resource):
    """A database user."""

    def __repr__(self):
        return "<User: %s>" % self.name


class Users(base.ManagerWithFind):
    """Manage :class:`Users` resources."""
    resource_class = User

    def create(self, instance, users):
        """Create users with permissions to the specified databases."""
        body = {"users": users}
        url = "/instances/%s/users" % base.getid(instance)
        resp, body = self.api.client.post(url, body=body)
        common.check_for_exceptions(resp, body, url)

    def delete(self, instance, username, hostname=None):
        """Delete an existing user in the specified instance."""
        user = common.quote_user_host(username, hostname)
        url = "/instances/%s/users/%s" % (base.getid(instance), user)
        resp, body = self.api.client.delete(url)
        common.check_for_exceptions(resp, body, url)

    def list(self, instance, limit=None, marker=None):
        """Get a list of all Users from the instance's Database.

        :rtype: list of :class:`User`.
        """
        url = "/instances/%s/users" % base.getid(instance)
        return self._paginated(url, "users", limit, marker)

    def get(self, instance, username, hostname=None):
        """Get a single User from the instance's Database.

        :rtype: :class:`User`.
        """
        user = common.quote_user_host(username, hostname)
        url = "/instances/%s/users/%s" % (base.getid(instance), user)
        return self._get(url, "user")

    def update_attributes(self, instance, username, newuserattr=None,
                          hostname=None):
        """Update attributes of a single User in an instance.

        :rtype: :class:`User`.
        """
        if not newuserattr:
            raise exceptions.ValidationError("No updates specified for user %s"
                                             % username)
        instance_id = base.getid(instance)
        user = common.quote_user_host(username, hostname)
        user_dict = {}
        user_dict['user'] = newuserattr
        url = "/instances/%s/users/%s" % (instance_id, user)
        resp, body = self.api.client.put(url, body=user_dict)
        common.check_for_exceptions(resp, body, url)

    def list_access(self, instance, username, hostname=None):
        """Show all databases the given user has access to."""
        instance_id = base.getid(instance)
        user = common.quote_user_host(username, hostname)
        url = "/instances/%(instance_id)s/users/%(user)s/databases"
        local_vars = locals()
        resp, body = self.api.client.get(url % local_vars)
        common.check_for_exceptions(resp, body, url)
        if not body:
            raise Exception("Call to %s did not return to a body" % url)
        return [databases.Database(self, db) for db in body['databases']]

    def grant(self, instance, username, databases, hostname=None):
        """Allow an existing user permissions to access a database."""
        instance_id = base.getid(instance)
        user = common.quote_user_host(username, hostname)
        url = "/instances/%(instance_id)s/users/%(user)s/databases"
        dbs = {'databases': [{'name': db} for db in databases]}
        local_vars = locals()
        resp, body = self.api.client.put(url % local_vars, body=dbs)
        common.check_for_exceptions(resp, body, url)

    def revoke(self, instance, username, database, hostname=None):
        """Revoke from an existing user access permissions to a database."""
        instance_id = base.getid(instance)
        user = common.quote_user_host(username, hostname)
        url = ("/instances/%(instance_id)s/users/%(user)s/"
               "databases/%(database)s")
        local_vars = locals()
        resp, body = self.api.client.delete(url % local_vars)
        common.check_for_exceptions(resp, body, url)

    def change_passwords(self, instance, users):
        """Change the password for one or more users."""
        instance_id = base.getid(instance)
        user_dict = {"users": users}
        url = "/instances/%s/users" % instance_id
        resp, body = self.api.client.put(url, body=user_dict)
        common.check_for_exceptions(resp, body, url)



# Copyright 2011 OpenStack Foundation
# Copyright 2013 Rackspace Hosting
# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools
from unittest import mock

from troveclient import common


class CommonTest(testtools.TestCase):

    def test_check_for_exceptions(self):
        status = [400, 422, 500]
        for s in status:
            resp = mock.Mock()
            resp.status_code = s
            self.assertRaises(Exception,
                              common.check_for_exceptions, resp, "body")

    def test_append_query_strings(self):
        url = "test-url"
        self.assertEqual(url, common.append_query_strings(url))

        self.assertEqual(url, common.append_query_strings(
            url, limit=None, marker=None))

        limit = "test-limit"
        marker = "test-marker"
        result = common.append_query_strings(url, limit=limit, marker=marker)
        self.assertTrue(result.startswith(url + '?'))
        self.assertIn("limit=%s" % limit, result)
        self.assertIn("marker=%s" % marker, result)
        self.assertEqual(1, result.count('&'))

        opts = {}
        self.assertEqual(url, common.append_query_strings(
            url, limit=None, marker=None, **opts))

        opts = {'key1': 'value1', 'key2': None}
        result = common.append_query_strings(url, limit=None, marker=marker,
                                             **opts)
        self.assertTrue(result.startswith(url + '?'))
        self.assertEqual(1, result.count('&'))
        self.assertNotIn("limit=%s" % limit, result)
        self.assertIn("marker=%s" % marker, result)
        self.assertIn("key1=%s" % opts['key1'], result)
        self.assertNotIn("key2=%s" % opts['key2'], result)

        opts = {'key1': 'value1', 'key2': None, 'key3': False}
        result = common.append_query_strings(url, **opts)
        self.assertTrue(result.startswith(url + '?'))
        self.assertIn("key1=value1", result)
        self.assertNotIn("key2", result)
        self.assertIn("key3=False", result)


class PaginatedTest(testtools.TestCase):

    def setUp(self):
        super(PaginatedTest, self).setUp()
        self.items_ = ["item1", "item2"]
        self.next_marker_ = "next-marker"
        self.links_ = ["link1", "link2"]
        self.pgn = common.Paginated(self.items_, self.next_marker_,
                                    self.links_)

    def tearDown(self):
        super(PaginatedTest, self).tearDown()

    def test___init__(self):
        self.assertEqual(self.items_, self.pgn)
        self.assertEqual(self.next_marker_, self.pgn.next)
        self.assertEqual(self.links_, self.pgn.links)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Database v1 Quota action implementations"""

from osc_lib.command import command
from osc_lib import utils as osc_utils
import six

from troveclient.i18n import _


class ShowDatabaseQuota(command.Lister):

    _description = _("Show quotas for a tenant.")
    columns = ['Resource', 'In Use', 'Reserved', 'Limit']

    def get_parser(self, prog_name):
        parser = super(ShowDatabaseQuota, self).get_parser(prog_name)
        parser.add_argument(
            'tenant_id',
            metavar='<tenant_id>',
            help=_('Id of tenant for which to show quotas.'),
        )
        return parser

    def take_action(self, parsed_args):
        db_quota = self.app.client_manager.database.quota
        quota = [osc_utils.get_item_properties(q, self.columns)
                 for q in db_quota.show(parsed_args.tenant_id)]
        return self.columns, quota


class UpdateDatabaseQuota(command.ShowOne):

    _description = _("Update quotas for a tenant.")

    def get_parser(self, prog_name):
        parser = super(UpdateDatabaseQuota, self).get_parser(prog_name)
        parser.add_argument(
            'tenant_id',
            metavar='<tenant_id>',
            help=_('Id of tenant for which to update quotas.'),
        )
        parser.add_argument(
            'resource',
            metavar='<resource>',
            help=_('Id of resource to change.'),
        )
        parser.add_argument(
            'limit',
            metavar='<limit>',
            type=int,
            help=_('New limit to set for the named resource.'),
        )
        return parser

    def take_action(self, parsed_args):
        db_quota = self.app.client_manager.database.quota
        update_params = {
            parsed_args.resource: parsed_args.limit
        }
        updated_quota = db_quota.update(parsed_args.tenant_id,
                                        update_params)
        return zip(*sorted(six.iteritems(updated_quota)))



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#


class FakeStdout(object):
    def __init__(self):
        self.content = []

    def write(self, text):
        self.content.append(text)

    def make_string(self):
        result = ''
        for line in self.content:
            result = result + line
        return result



# Copyright 2011 OpenStack Foundation
# Copyright 2013 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from troveclient import base
from troveclient import common


class Account(base.Resource):
    """Account is an opaque instance used to hold account information."""
    def __repr__(self):
        return "<Account: %s>" % self.name


class Accounts(base.ManagerWithFind):
    """Manage :class:`Account` information."""

    resource_class = Account

    def _list(self, url, response_key):
        resp, body = self.api.client.get(url)
        if not body:
            raise Exception("Call to " + url + " did not return a body.")
        return self.resource_class(self, body[response_key])

    def index(self):
        """Get a list of all accounts with non-deleted instances."""

        url = "/mgmt/accounts"
        resp, body = self.api.client.get(url)
        common.check_for_exceptions(resp, body, url)
        if not body:
            raise Exception("Call to " + url + " did not return a body.")
        return base.Resource(self, body)

    def show(self, account):
        """Get details of one account.

        :rtype: :class:`Account`.
        """

        acct_name = self._get_account_name(account)
        return self._list("/mgmt/accounts/%s" % acct_name, 'account')

    # Appease the abc gods
    def list(self):
        pass

    @staticmethod
    def _get_account_name(account):
        try:
            if account.name:
                return account.name
        except AttributeError:
            return account



# Copyright 2011 OpenStack Foundation
# Copyright 2013 Rackspace Hosting
# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools
from unittest import mock

from troveclient import base
from troveclient.v1 import instances

"""
Unit tests for instances.py
"""


class InstanceTest(testtools.TestCase):

    def setUp(self):
        super(InstanceTest, self).setUp()
        self.orig__init = instances.Instance.__init__
        instances.Instance.__init__ = mock.Mock(return_value=None)
        self.instance = instances.Instance()
        self.instance.manager = mock.Mock()

    def tearDown(self):
        super(InstanceTest, self).tearDown()
        instances.Instance.__init__ = self.orig__init

    def test_list_databases(self):
        db_list = ['database1', 'database2']
        self.instance.manager.databases = mock.Mock()
        self.instance.manager.databases.list = mock.Mock(return_value=db_list)
        self.assertEqual(db_list, self.instance.list_databases())

    def test_delete(self):
        db_delete_mock = mock.Mock(return_value=None)
        self.instance.manager.delete = db_delete_mock
        self.instance.delete()
        self.assertEqual(1, db_delete_mock.call_count)

    def test_restart(self):
        db_restart_mock = mock.Mock(return_value=None)
        self.instance.manager.restart = db_restart_mock
        self.instance.id = 1
        self.instance.restart()
        self.assertEqual(1, db_restart_mock.call_count)

    def test_detach_replica(self):
        db_detach_mock = mock.Mock(return_value=None)
        self.instance.manager.edit = db_detach_mock
        self.instance.id = 1
        self.instance.detach_replica()
        self.assertEqual(1, db_detach_mock.call_count)


class InstancesTest(testtools.TestCase):

    def setUp(self):
        super(InstancesTest, self).setUp()
        self.orig__init = instances.Instances.__init__
        instances.Instances.__init__ = mock.Mock(return_value=None)
        self.instances = instances.Instances()
        self.instances.api = mock.Mock()
        self.instances.api.client = mock.Mock()
        self.instances.resource_class = mock.Mock(return_value="instance-1")

        self.instance_with_id = mock.Mock()
        self.instance_with_id.id = 215

    def tearDown(self):
        super(InstancesTest, self).tearDown()
        instances.Instances.__init__ = self.orig__init

    def test_create(self):
        def side_effect_func(path, body, inst):
            return path, body, inst

        self.instances._create = mock.Mock(side_effect=side_effect_func)
        nics = [{'net-id': '000'}]
        p, b, i = self.instances.create("test-name", 103, "test-volume",
                                        ['db1', 'db2'], ['u1', 'u2'],
                                        datastore="datastore",
                                        datastore_version="datastore-version",
                                        nics=nics, replica_of='test',
                                        replica_count=4,
                                        modules=['mod_id'],
                                        locality='affinity',
                                        access={'is_public': True})
        self.assertEqual("/instances", p)
        self.assertEqual("instance", i)
        self.assertEqual(['db1', 'db2'], b["instance"]["databases"])
        self.assertEqual(['u1', 'u2'], b["instance"]["users"])
        self.assertEqual("test-name", b["instance"]["name"])
        self.assertEqual("test-volume", b["instance"]["volume"])
        self.assertEqual("datastore", b["instance"]["datastore"]["type"])
        self.assertEqual("datastore-version",
                         b["instance"]["datastore"]["version"])
        self.assertEqual(nics, b["instance"]["nics"])
        self.assertEqual(103, b["instance"]["flavorRef"])
        self.assertEqual('test', b['instance']['replica_of'])
        self.assertEqual([{'id': 'mod_id'}], b["instance"]["modules"])
        self.assertEqual(4, b["instance"]["replica_count"])
        self.assertEqual('affinity', b["instance"]["locality"])
        self.assertEqual({'is_public': True}, b["instance"]["access"])

    def test_list(self):
        page_mock = mock.Mock()
        self.instances._paginated = page_mock
        limit = "test-limit"
        marker = "test-marker"
        include_clustered = {'include_clustered': False}
        self.instances.list(limit, marker)
        page_mock.assert_called_with("/instances", "instances", limit, marker,
                                     include_clustered)

    def test_detailed_list(self):
        page_mock = mock.Mock()
        self.instances._paginated = page_mock
        limit = "test-limit"
        marker = "test-marker"
        include_clustered = {'include_clustered': False}
        self.instances.list(limit, marker, detailed=True)
        page_mock.assert_called_with("/instances/detail", "instances", limit,
                                     marker, include_clustered)

    def test_get(self):
        def side_effect_func(path, inst):
            return path, inst

        self.instances._get = mock.Mock(side_effect=side_effect_func)
        self.assertEqual(('/instances/instance1', 'instance'),
                         self.instances.get('instance1'))

    def test_delete(self):
        resp = mock.Mock()
        resp.status_code = 200
        body = None
        self.instances.api.client.delete = mock.Mock(return_value=(resp, body))
        self.instances.delete('instance1')
        self.instances.delete(self.instance_with_id)
        resp.status_code = 500
        self.assertRaises(Exception, self.instances.delete, 'instance1')

    def test__action(self):
        body = mock.Mock()
        resp = mock.Mock()
        resp.status_code = 200
        self.instances.api.client.post = mock.Mock(return_value=(resp, body))
        self.assertEqual('instance-1', self.instances._action(1, body))

        self.instances.api.client.post = mock.Mock(return_value=(resp, None))
        self.assertIsNone(self.instances._action(1, body))

    def _set_action_mock(self):
        def side_effect_func(instance, body):
            self._instance_id = base.getid(instance)
            self._body = body

        self._instance_id = None
        self._body = None
        self.instances._action = mock.Mock(side_effect=side_effect_func)

    def _test_resize_volume(self, instance, id):
        self._set_action_mock()
        self.instances.resize_volume(instance, 1024)
        self.assertEqual(id, self._instance_id)
        self.assertEqual({"resize": {"volume": {"size": 1024}}}, self._body)

    def test_resize_volume_with_id(self):
        self._test_resize_volume(152, 152)

    def test_resize_volume_with_obj(self):
        self._test_resize_volume(self.instance_with_id,
                                 self.instance_with_id.id)

    def _test_resize_instance(self, instance, id):
        self._set_action_mock()
        self.instances.resize_instance(instance, 103)
        self.assertEqual(id, self._instance_id)
        self.assertEqual({"resize": {"flavorRef": 103}}, self._body)

    def test_resize_instance_with_id(self):
        self._test_resize_instance(4725, 4725)

    def test_resize_instance_with_obj(self):
        self._test_resize_instance(self.instance_with_id,
                                   self.instance_with_id.id)

    def _test_restart(self, instance, id):
        self._set_action_mock()
        self.instances.restart(instance)
        self.assertEqual(id, self._instance_id)
        self.assertEqual({'restart': {}}, self._body)

    def test_restart_with_id(self):
        self._test_restart(253, 253)

    def test_restart_with_obj(self):
        self._test_restart(self.instance_with_id, self.instance_with_id.id)

    def test_modify(self):
        resp = mock.Mock()
        resp.status_code = 200
        body = None
        self.instances.api.client.put = mock.Mock(return_value=(resp, body))
        self.instances.modify(123)
        self.instances.modify(123, 321)
        self.instances.modify(self.instance_with_id)
        self.instances.modify(self.instance_with_id, 123)
        resp.status_code = 500
        self.assertRaises(Exception, self.instances.modify, 'instance1')

    def test_edit(self):
        resp = mock.Mock()
        resp.status_code = 204

        def fake_patch(url, body):
            # Make sure we never pass slave_of to the API.
            self.assertIn('instance', body)
            self.assertNotIn('slave_of', body['instance'])
            return resp, None

        self.instances.api.client.patch = mock.Mock(side_effect=fake_patch)
        self.instances.edit(123)
        self.instances.edit(123, 321)
        self.instances.edit(123, 321, 'name-1234')
        self.instances.edit(123, 321, 'name-1234', True)
        self.instances.edit(self.instance_with_id)
        self.instances.edit(self.instance_with_id, 123)
        self.instances.edit(self.instance_with_id, 123, 'name-1234')
        self.instances.edit(self.instance_with_id, 123, 'name-1234', True)
        resp.status_code = 500
        self.assertRaises(Exception, self.instances.edit, 'instance1')

    def test_module_apply(self):
        resp = mock.Mock()
        resp.status_code = 200
        body = {'modules': []}
        self.instances.api.client.post = mock.Mock(return_value=(resp, body))
        self.instances.module_apply(self.instance_with_id, "mod_id")
        resp.status_code = 500
        self.assertRaises(Exception, self.instances.module_apply,
                          'instance1', 'mod1')

    def test_module_remove(self):
        resp = mock.Mock()
        resp.status_code = 200
        body = {'modules': []}
        self.instances.api.client.delete = mock.Mock(return_value=(resp, body))
        self.instances.module_remove(self.instance_with_id, "mod_id")
        resp.status_code = 500
        self.assertRaises(Exception, self.instances.module_remove,
                          'instance1', 'mod1')

    def test_module_query(self):
        resp = mock.Mock()
        resp.status_code = 200
        body = {'modules': []}
        self.instances.api.client.get = mock.Mock(return_value=(resp, body))
        self.instances.module_query(self.instance_with_id)
        resp.status_code = 500
        self.assertRaises(Exception, self.instances.module_query,
                          'instance1')

    def test_module_retrieve(self):
        resp = mock.Mock()
        resp.status_code = 200
        body = {'modules': []}
        self.instances.api.client.get = mock.Mock(return_value=(resp, body))
        self.instances.module_retrieve(self.instance_with_id)
        resp.status_code = 500
        self.assertRaises(Exception, self.instances.module_retrieve,
                          'instance1')

    def test_module_list_instance(self):
        resp = mock.Mock()
        resp.status_code = 200
        body = {'modules': []}
        self.instances.api.client.get = mock.Mock(return_value=(resp, body))
        self.instances.modules(self.instance_with_id)
        resp.status_code = 500
        self.assertRaises(Exception, self.instances.modules,
                          'instance1')

    def test_upgrade(self):
        resp = mock.Mock()
        resp.status_code = 200
        body = None
        self.instances.api.client.patch = mock.Mock(return_value=(resp, body))
        self.instances.upgrade(self.instance_with_id, "5.6")
        resp.status_code = 500
        self.assertRaises(Exception, self.instances.upgrade,
                          'instance1')

    def test_configuration(self):
        def side_effect_func(path, inst):
            return path, inst

        self.instances._get = mock.Mock(side_effect=side_effect_func)
        self.assertEqual(('/instances/instance1/configuration', 'instance'),
                         self.instances.configuration('instance1'))


class InstanceStatusTest(testtools.TestCase):

    def test_constants(self):
        self.assertEqual("ACTIVE", instances.InstanceStatus.ACTIVE)
        self.assertEqual("BLOCKED", instances.InstanceStatus.BLOCKED)
        self.assertEqual("BUILD", instances.InstanceStatus.BUILD)
        self.assertEqual("FAILED", instances.InstanceStatus.FAILED)
        self.assertEqual("REBOOT", instances.InstanceStatus.REBOOT)
        self.assertEqual("RESIZE", instances.InstanceStatus.RESIZE)
        self.assertEqual("SHUTDOWN", instances.InstanceStatus.SHUTDOWN)
        self.assertEqual("RESTART_REQUIRED",
                         instances.InstanceStatus.RESTART_REQUIRED)
